import pprint
import sys

from depcheck.models import Ruleset, DependencyReport


class DepChecker:
    __ruleset: Ruleset
    __dependency_report: DependencyReport

    def __init__(self, ruleset: Ruleset,
                 dependency_report: DependencyReport) -> None:
        self.__ruleset = ruleset
        self.__dependency_report = dependency_report

    def run(self) -> None:
        # Check dependencies if there is any violation for given ruleset
        errors = {}
        for layer in self.__ruleset.rules:
            violations = self.__check_rule(layer)
            if len(violations) > 0:
                errors[layer] = violations

        if errors:
            print(f"There are {len(errors)} forbidden dependencies:")
            pprint.pprint(errors)
            sys.exit(1)
        else:
            print("OK")
            sys.exit(0)

    def __check_rule(self, layer: str) -> list[str]:
        violation: list[str] = []
        layer_deps: set[str] = self.__dependency_report.layer_dependencies(layer)
        whitelist: set[str] = self.__dependency_report.whitelist(layer)

        # Since some packages depend on the root package but the whitelist doesn't have
        # the root package explicitly, we need to add it to the whitelist.
        whitelist.add(self.__dependency_report.root_package)

        black_list = layer_deps.difference(whitelist)

        if len(black_list) > 0:
            violation.extend(list(black_list))

        return violation



import json
from typing import Any

import yaml
from pytest import raises, fixture

from depcheck.depchecker import DepChecker
from depcheck.models import Ruleset, DependencyReport


@fixture()
def ruleset() -> Ruleset:
    """
    This is a simple ruleset which includes two layers which are layer_1 and layer_2.
    - Depended modules of each layer take place in the 'layers' part.
        e.g. layer_1 is depending on the 'root_package.package_1' module.
    - Dependencies between layers are configured in the 'whitelist' part. e.g:
        - 'layer_1: ~': layer_1 should not depend on any layers
        - 'layer_2: layer_1': layer_2 is allowed to depend on layer_1
    """
    config_yaml = '''
            layers:
              layer_1:
                - root_package.package_1
              layer_2:
              - root_package.package_2
            whitelist:
              layer_1: ~
              layer_2:
                - layer_1
    '''

    ruleset = yaml.safe_load(config_yaml)

    return Ruleset(layers=ruleset['layers'], rules=ruleset['whitelist'])


@fixture
def valid_dependencies() -> dict[str, Any]:
    """
    - 'root_package.package_1' module inside 'layer_1' doesn't import any modules.
    i.e. layer_1 doesn't depend on any layers.
    - 'root_package.package_2' module inside 'layer_2' depends on the 'root_package.package_1'
    module inside 'layer_1'. i.e. layer_2 depends on layer_1
    """
    return json.loads('''
                {
                    "__main__": {
                        "bacon": 0,
                        "imports": [
                            "root_package",
                            "root_package.__main__",
                            "root_package.package_1",
                            "root_package.package_1.model",
                            "root_package.package_2",
                            "root_package.package_2.usecase"
                        ],
                        "name": "__main__"
                    },
                    "root_package": {
                        "bacon": 1,
                        "imports": [
                            "root_package",
                            "root_package.package_1",
                            "root_package.package_2"
                        ],
                        "name": "root_package"
                    },
                    "root_package.__main__": {
                        "bacon": 1,
                        "name": "root_package.__main__"
                    },
                    "root_package.package_1": {
                        "bacon": 1,
                        "name": "root_package.package_1"
                    },
                    "root_package.package_1.model": {
                        "bacon": 1,
                        "name": "root_package.package_1.model"
                    },
                    "root_package.package_2": {
                        "bacon": 1,
                        "name": "root_package.package_2"
                    },
                    "root_package.package_2.usecase": {
                        "bacon": 1,
                        "imports": [
                            "root_package",
                            "root_package.package_1",
                            "root_package.package_1.model"
                        ],
                        "name": "root_package.package_2.usecase"
                    }
                }

        ''')


@fixture
def forbidden_dependencies() -> dict[str, Any]:
    """
        - 'root_package.package_1' module inside 'layer_1' depends on the 'root_package.package_2'
        module inside 'layer_2'. i.e. layer_1 depends on layer_2
        - 'root_package.package_2' module inside 'layer_2' depends on the 'root_package.package_1'
        module inside 'layer_1'. i.e. layer_2 depends on layer_1
        """
    return json.loads('''
                {
                    "__main__": {
                        "imports": [
                            "root_package",
                            "root_package.__main__",
                            "root_package.package_1",
                            "root_package.package_1.model",
                            "root_package.package_2",
                            "root_package.package_2.usecase"
                        ],
                        "name": "__main__"
                    },
                    "root_package": {
                        "imports": [
                            "root_package",
                            "root_package.package_1",
                            "root_package.package_2"
                        ],
                        "name": "root_package"
                    },
                    "root_package.package_1": {
                        "bacon": 1,
                        "name": "root_package.package_1"
                    },
                    "root_package.package_1.model": {
                        "bacon": 1,
                        "name": "root_package.package_1.model"
                    },
                    "root_package.package_2": {
                        "bacon": 1,
                        "name": "root_package.package_2"
                    },
                    "root_package.package_1.model": {
                        "imports": [
                            "root_package.package_2.usecase"
                        ],
                        "name": "root_package.package_1.model"
                    },
                    "root_package.package_2.usecase": {
                        "imports": [
                            "root_package",
                            "root_package.package_1",
                            "root_package.package_1.model"
                        ],
                        "name": "root_package.package_2.usecase"
                    }
                }
        ''')


def test_project_dependencies_are_valid(ruleset: Ruleset,
                                        valid_dependencies: dict[str, Any]) -> None:
    dependency_report: DependencyReport = DependencyReport(ruleset, valid_dependencies)

    with raises(SystemExit) as exception:
        DepChecker(ruleset, dependency_report).run()

    assert exception.type == SystemExit
    assert exception.value.code == 0


def test_project_dependencies_are_forbidden(ruleset: Ruleset,
                                            forbidden_dependencies: dict[str, Any]) -> None:
    dependency_report: DependencyReport = DependencyReport(ruleset, forbidden_dependencies)

    with raises(SystemExit) as exception:
        DepChecker(ruleset, dependency_report).run()

    assert exception.type == SystemExit
    assert exception.value.code == 1



from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from depcheck.depchecker import DepChecker
from depcheck.models import Ruleset, DependencyReport
from depcheck.readers import RulesetReader, DependencyReader

DEFAULT_CONFIG_FILE_PATH = ".depcheck.yml"


def main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    add_arguments(parser)
    args = parser.parse_args()

    print("# Reading ruleset...")
    ruleset: Ruleset = RulesetReader(args.file).read()
    print("# Reading project packages...")
    dependency_report: DependencyReport = DependencyReader(ruleset, args.root_package).read()
    print("# Checking dependencies...")
    DepChecker(ruleset, dependency_report).run()


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "root_package",
        help="Root package of the project"
    )

    parser.add_argument(
        "-f",
        "--file",
        default=DEFAULT_CONFIG_FILE_PATH,
        help="Path to the config file that includes layers and rules.",
    )


if __name__ == "__main__":
    main()



import json
import subprocess
from typing import Any

import yaml

from depcheck.models import Ruleset, DependencyReport


class RulesetReader:
    __filename: str

    def __init__(self, filename: str) -> None:
        self.__filename = filename

    def read(self) -> Ruleset:
        ruleset: dict[str, Any]

        with open(self.__filename, 'r') as stream:
            try:
                ruleset = yaml.safe_load(stream)
            except yaml.YAMLError:
                print('\n\n[!] Command must be run from project root '
                      'and .depcheck.yml file should be in the root\n\n')

        return Ruleset(layers=ruleset['layers'], rules=ruleset['whitelist'])


class DependencyReader:
    __ruleset: Ruleset
    __root_package: str

    def __init__(self, ruleset: Ruleset, root_package: str) -> None:
        self.__ruleset = ruleset
        self.__root_package = root_package

    def read(self) -> DependencyReport:
        output = subprocess.check_output("pydeps --show-deps --no-show --no-output --max-bacon 2 "
                                         + self.__root_package,
                                         shell=True).decode('u8').strip()
        dependency_graph: dict[str, Any] = json.loads(output)

        return DependencyReport(self.__ruleset, dependency_graph)



from depcheck.util import flatten


def test_flatten_returns_items_of_inner_lists_in_a_flatten_list() -> None:
    # Arrange
    list_2d: list[list[str]] = [['a', 'b', 'c'],
                                ['1', '2'],
                                ['x', 'y', 'z']]
    expected_flatten_list: list[str] = ['a', 'b', 'c', '1', '2', 'x', 'y', 'z']

    # Act
    result: list[str] = flatten(list_2d)

    # Assert
    assert expected_flatten_list == result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['depcheck']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0', 'pydeps>=1.9.13,<2.0.0']

entry_points = \
{'console_scripts': ['depcheck = depcheck.main:main']}

setup_kwargs = {
    'name': 'depcheck',
    'version': '0.1.1',
    'description': 'Depcheck is a tool to check package dependencies between predefined layers to make sure that the application always complies with the Hexagonal Architecture principle of creating loosely coupled application components.',
    'long_description': "![Depcheck: Dependency Checker](/docs/.img/depcheck_logo.jpg)\n\nDepcheck is a tool to check package-dependencies between predefined layers. \nIn the configuration file(`.depcheck.yml`) located in the project root, \nwhich packages belong to which layers and allowed dependencies between \nlayers are configurable. In this way, you can make sure that the application \nalways complies with the [Hexagonal Architecture][hexagonal-architecture] \nprinciple of creating loosely coupled application components.\n\n## Install\nInstall from [Pypi][pypi-link] via `pip install depcheck`\n    \n## Usage\nLet's say you have a project with the directory structure below:\n```text\nproject_directory\n    root_package\n        package-1\n        package-2\n        main.py\n    README.md\n    .gitignore\n    .depcheck.yml\n```\n- Navigate to the `project_directory` then run `depcheck` for your project:\n    ```shell\n    depcheck root_package\n    ```\n- As you can see in the directory structure above, we have `.depcheck.yml` \n  configuration file in the project directory. If you would like to change \n  the path of the configuration file, use `-f` or `--file` argument:\n    ```shell\n    depcheck root_package -f config/customized_depcheck.yml\n    ```\n\n## Contributing\nAll contributions are welcomed! Start by searching through the [issues][issues] and \n[pull requests][pull-requests] to see whether someone else has raised a similar idea or question.\nIf you don't see your idea listed, open a pull request.\n\n## Maintenance Work\n- [Upgrade project's Python version][upgrade-python-version]\n- [Update project dependencies][update-project-dependencies]\n\n\n<!-- Links -->\n[hexagonal-architecture]: https://en.wikipedia.org/wiki/Hexagonal_architecture_(software)\n[upgrade-python-version]: ./docs/upgrade-python-version.md\n[update-project-dependencies]: ./docs/upgrade-python-version.md\n[pypi-link]: https://pypi.org/project/depcheck/\n[issues]: https://github.com/flix-tech/depcheck/issues\n[pull-requests]: https://github.com/flix-tech/depcheck/pulls\n",
    'author': 'FlixMobility Tech',
    'author_email': 'open-source@flixbus.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/flix-tech/depcheck',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.6,<4.0.0',
}


setup(**setup_kwargs)



from typing import Any

from depcheck.util import flatten


class Ruleset:
    __rules: dict[str, list[str]]
    __layers: dict[str, list[str]]

    def __init__(self, layers: dict[str, list[str]],
                 rules: dict[str, list[str]]) -> None:
        self.__rules = rules
        self.__layers = layers

    def layer_packages(self, layer: str) -> list[str]:
        return self.layers.get(layer, [])

    def whitelist(self, layer: str) -> list[str]:
        return self.rules.get(layer, [])

    @property
    def rules(self) -> dict[str, list[str]]:
        return self.__rules

    @property
    def layers(self) -> dict[str, list[str]]:
        return self.__layers


class DependencyReport:
    __ruleset: Ruleset
    __graph: dict[str, Any]
    __all_packages: list[str]

    def __init__(self, ruleset: Ruleset,
                 dependency_graph: dict[str, Any]) -> None:
        self.__ruleset = ruleset
        self.__graph = dependency_graph
        self.__all_packages = list(self.__graph.keys())

    def layer_dependencies(self, layer: str) -> set[str]:
        layer_packages: set[str] = self.__layer_packages([layer])
        layer_dependencies = self.__package_dependencies(layer_packages)

        return self.__exclude_inner_dependencies(layer_dependencies, layer_packages)

    def whitelist(self, layer: str) -> set[str]:
        whitelist: list[str] = self.__ruleset.whitelist(layer)

        return self.__layer_packages(whitelist)

    @property
    def root_package(self) -> str:
        return self.__graph['__main__']['imports'][0]

    @staticmethod
    def __exclude_inner_dependencies(raw_dependencies: set[str], layers: set[str]) -> set[str]:
        return raw_dependencies - layers

    def __layer_packages(self, layers: list[str]) -> set[str]:
        if layers is None:
            return set()

        # Packages within given layers
        layer_packages: list[str] = flatten([self.__ruleset.layer_packages(layer)
                                             for layer in layers])

        # Include subpackages of layer_packages
        layer_packages.extend(self.__subpackages(layer_packages))

        return set(layer_packages)

    def __subpackages(self, packages: list[str]) -> list[str]:
        subpackages: list[str] = []

        for package in self.__all_packages:
            if package.startswith(tuple(packages)):
                subpackages.append(package)

        return subpackages

    def __package_dependencies(self, packages: set[str]) -> set[str]:
        """
        Returns all packages that are used(imported) in given 'packages'.
        In other words, this returns dependencies of given 'packages'.
        """
        return set(flatten([self.__graph.get(i, {}).get('imports', []) for i in packages]))
